/*
 * Generation of message structure access operations.
 *
 * Copyright (C) 2019, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "common.h"
#include "message.h"
#include "templates.h"

/* Generate a variable declaration for message word accessors. */

void write_accessor_declaration(int input_words, int output_words,
                                const char *opname, FILE *fp)
{
  if (input_words)
    fprintf(fp, message_accessor_declaration, "in", opname, "in");

  if (output_words)
    fprintf(fp, message_accessor_declaration, "out", opname, "out");
}

/* Generate a variable initialisation for message word accessors. */

void write_accessor_initialisation(enum specifier direction,
                                   enum component_role component,
                                   enum function_role function,
                                   const char *opname, FILE *fp)
{
  char *prefix = structure_prefix(direction);
  char *addr = reference_message(component, function);

  fprintf(fp, message_accessor_initialisation, prefix, prefix, opname);

  /* When writing to the message, the accessor will reserve space.
     When reading from the message, only words will be present. */

  if (writing_to_message(component, direction))
    fprintf(fp, message_accessor_writing_initialiser, addr, prefix, opname);
  else
    fprintf(fp, message_accessor_reading_initialiser, addr);
}

/* Copy values between a message and local names. */

void write_message_access(struct parameter *param, enum component_role component,
                          enum function_role function, enum specifier direction,
                          enum parameter_class cls, FILE *fp)
{
  char *name, *access, *prefix, *suffix, *addr;
  int writing = writing_to_message(component, direction);
  int index;

  /* Establish type and variable naming or message access conventions. */

  switch (cls)
  {
    case WORD_CLASS: prefix = structure_prefix(direction); break;
    case ITEM_CLASS: addr = reference_message(component, function); break;
  }

  /* Generate an access operation for parameters having the indicated direction,
     and having the indicated class (word or item). */

  for (index = 0; param != NULL; param = param->tail)
  {
    /* Permit the specification of receive window flexpage input items for
       "out fpage" parameters. */

    if ((param->cls & cls) &&
        ((param->specifier & direction) ||
         specify_receive_fpage(param, component, writing)))
    {
      name = get_parameter_name(param->identifiers);
      access = access_name(param, component, writing);

      switch (cls)
      {
        /* For words, access the message structure. */

        case WORD_CLASS:
          if (writing)
            fprintf(fp, "  %s_words->%s = %s%s;\n", prefix, name, access, name);
          else
            fprintf(fp, "  %s%s = %s_words->%s;\n", access, name, prefix, name);
          break;

        /* For items, perform access operations. */

        case ITEM_CLASS:
          suffix = item_type_name(param, component, writing);

          if (writing)
            fprintf(fp, "  ipc_message_add_%s(%smsg, %s%s);\n",
                    suffix, addr, access, name);
          else
            fprintf(fp, "  ipc_message_import_%s(%smsg, %d, %s%s);\n",
                    suffix, addr, index, access, name);
          index++;
          break;

        default:
          break;
      }
    }
  }
}
